@@ -1,7 +1,8 @@
 #!/usr/bin/python3
 # -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
 #   MenuLibre - Advanced fd.o Compliant Menu Editor
-#   Copyright (C) 2012-2015 Sean Davis <smd.seandavis@gmail.com>
+#   Copyright (C) 2012-2022 Sean Davis <sean@bluesabre.org>
+#   Copyright (C) 2016 OmegaPhil <OmegaPhil@startmail.com>
 #
 #   This program is free software: you can redistribute it and/or modify it
 #   under the terms of the GNU General Public License version 3, as published
@@ -16,31 +17,43 @@
 #   with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import os
+import subprocess
+
 from locale import gettext as _
 
 from gi.repository import Gio, GObject, Gtk, Pango, GLib
 
 from . import MenuEditor, MenulibreXdg, XmlMenuElementTree, util
-from .util import MenuItemTypes, check_keypress, getBasename
+from .util import MenuItemTypes, check_keypress, getBasename, getRelativeName, escapeText
 
 import logging
 logger = logging.getLogger('menulibre')
 
+
 class Treeview(GObject.GObject):
 
     __gsignals__ = {
         'cursor-changed': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_BOOLEAN,
-                        (GObject.TYPE_BOOLEAN,)),
-        'add-directory-enabled': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_BOOLEAN,
-                        (GObject.TYPE_BOOLEAN,)),
+                           (GObject.TYPE_BOOLEAN,)),
+        'add-directory-enabled': (GObject.SIGNAL_RUN_LAST,
+                                  GObject.TYPE_BOOLEAN,
+                                  (GObject.TYPE_BOOLEAN,)),
+        'requires-menu-reload': (GObject.SIGNAL_RUN_LAST,
+                                  GObject.TYPE_BOOLEAN,
+                                  (GObject.TYPE_BOOLEAN,)),
     }
 
+    loaded = False
+
     def __init__(self, parent, builder):
         GObject.GObject.__init__(self)
         self.parent = parent
 
         # Configure Widgets
-        self._configure_treeview(builder)
+        if self._configure_treeview(builder):
+            self.loaded = True
+        else:
+            return
         self._configure_toolbar(builder)
 
         # Defaults
@@ -52,10 +65,12 @@ class Treeview(GObject.GObject):
         """Configure the TreeView widget."""
         # Get the menu treestore.
         treestore = MenuEditor.get_treestore()
+        if not treestore:
+            return False
 
         self._treeview = builder.get_object('classic_view_treeview')
 
-        # Create a new column.
+        # Translators: "Search Results" treeview column header
         col = Gtk.TreeViewColumn(_("Search Results"))
 
         # Create and pack the PixbufRenderer.
@@ -71,6 +86,9 @@ class Treeview(GObject.GObject):
         # Set the markup property on the Text cell.
         col.add_attribute(col_cell_text, "markup", 0)
 
+        # Add the cell data func for the pixbuf column to render icons.
+        col.set_cell_data_func(col_cell_text, self._text_display_func, None)
+
         # Set the Tooltip column.
         self._treeview.set_tooltip_column(1)
 
@@ -83,18 +101,22 @@ class Treeview(GObject.GObject):
 
         # Configure the treeview events.
         self._treeview.connect("cursor-changed",
-                                self._on_treeview_cursor_changed, None, builder)
+                               self._on_treeview_cursor_changed, None, builder)
         self._treeview.connect("key-press-event",
-                                self._on_treeview_key_press_event, None)
+                               self._on_treeview_key_press_event, None)
         self._treeview.connect("row-expanded",
-                                self._on_treeview_row_expansion, True)
+                               self._on_treeview_row_expansion, True)
         self._treeview.connect("row-collapsed",
-                                self._on_treeview_row_expansion, False)
+                               self._on_treeview_row_expansion, False)
 
         # Show the treeview, grab focus.
         self._treeview.show_all()
         self._treeview.grab_focus()
 
+        self.menu_timeout_id = 0
+
+        return True
+
     def _configure_toolbar(self, builder):
         """Configure the toolbar widget."""
         self._toolbar = builder.get_object('browser_toolbar')
@@ -102,6 +124,20 @@ class Treeview(GObject.GObject):
         move_up.connect('clicked', self._move_iter, (self._treeview, -1))
         move_down = builder.get_object('classic_view_move_down')
         move_down.connect('clicked', self._move_iter, (self._treeview, 1))
+        sort = builder.get_object('classic_view_sort')
+        sort.connect('clicked', self._sort_iter)
+        self._sort_button = sort
+        self._move_up_button = move_up
+        self._move_down_button = move_down
+
+    def set_sortable(self, sortable):
+        self._sort_button.set_sensitive(sortable)
+
+    def set_move_up_enabled(self, enabled):
+        self._move_up_button.set_sensitive(enabled)
+
+    def set_move_down_enabled(self, enabled):
+        self._move_down_button.set_sensitive(enabled)
 
 # TreeView Modifiers
     def append(self, row_data):
@@ -124,82 +160,108 @@ class Treeview(GObject.GObject):
 
         return new_iter
 
-    def add_child(self, row_data):
-        """Add a new child launcher to the current selected one."""
-        model, treeiter = self._get_selected_iter()
+    def add_child(self, row_data, treeiter=None, model=None, do_select=True):
+        """Add a new child launcher to the current selected one, or the
+        specified iter if adding elsewhere in the tree, with optional
+        selection."""
+        if treeiter is None or model is None:
+            model, treeiter = self._get_selected_iter()
 
         new_iter = model.prepend(treeiter)
-        self._treeview.expand_row(model[treeiter].path, False)
 
-        self._populate_and_select_iter(model, new_iter, row_data)
+        if do_select:
+            self._treeview.expand_row(model[treeiter].path, False)
+
+        self._populate_and_select_iter(model, new_iter, row_data, do_select)
 
         return new_iter
 
-    def remove_selected(self):
-        """Remove the selected launcher."""
+    def remove_selected(self, ui_only=False):  # noqa
+        """Remove the selected launcher. If ui_only is True, the associated
+        desktop file is not deleted but the launcher is removed from the
+        interface. This is useful for removing categories."""
+
         self._last_selected_path = -1
         model, treeiter = self._get_selected_iter()
 
-        filename = model[treeiter][5]
-        item_type = model[treeiter][2]
-        if filename is not None:
-            basename = getBasename(filename)
-            original = util.getSystemLauncherPath(basename)
-        else:
-            original = None
+        if not ui_only:
 
-        # Get files for deletion
-        del_dirs, del_apps = self._get_delete_filenames(model, treeiter)
-        del_files = del_dirs + del_apps
+            filename = model[treeiter][MenuEditor.COL_FILENAME]
+            item_type = model[treeiter][MenuEditor.COL_TYPE]
+            if filename is not None:
+                basename = getRelativeName(filename)
+                original = util.getSystemLauncherPath(basename)
+            else:
+                original = None
 
-        # Uninstall the launcher
-        self.xdg_menu_uninstall(model, treeiter, filename)
+            # Get files for deletion - only one item can be selected at a time,
+            # but this may be a directory
+            del_dirs, del_apps = self._get_delete_filenames(model, treeiter)
+            del_files = del_dirs + del_apps
 
-        # Delete each of the files
-        for filename in del_files:
-            try:
-                os.remove(filename)
-            except:
-                pass
+            # Uninstall the launcher
+            self.xdg_menu_uninstall(model, treeiter, filename)
+
+            # Delete each of the files - this will fail silently for non-user
+            # desktop files/directories, these are hidden below
+            for filename in del_files:
+                try:
+                    os.remove(filename)
+                except:  # noqa
+                    pass
 
         self.xdg_menu_update()
         self._cleanup_applications_merged()
 
-        if filename not in del_files:
-            # Update the required categories.
-            model, parent_data = self.get_parent_row_data()
-            if parent_data is not None:
-                categories = util.getRequiredCategories(parent_data[5])
-            else:
-                categories = util.getRequiredCategories(None)
-            self.parent.update_launcher_categories(categories, [])
+        if not ui_only:
 
-        if original is not None:
-            # Original found, replace.
-            entry = MenulibreXdg.MenulibreDesktopEntry(original)
-            name = entry['Name']
-            comment = entry['Comment']
-            icon_name = entry['Icon']
-            if os.path.isfile(icon_name):
-                gfile = Gio.File.parse_name(icon_name)
-                icon = Gio.FileIcon.new(gfile)
-            else:
-                icon = Gio.ThemedIcon.new(icon_name)
-            self.update_selected(name, comment, item_type, icon_name,
-                                         original)
-            model, row_data = self.get_selected_row_data()
-            self.update_launcher_instances(filename, row_data)
-            treeiter = None
+            # How is it possible that a launcher is to be removed, but its
+            # associated desktop file has not been returned to be deleted?
+            if filename not in del_files:
+                # Update the required categories.
+                model, parent_data = self.get_parent_row_data()
+                if parent_data is not None:
+                    categories = util.getRequiredCategories(parent_data[MenuEditor.COL_FILENAME])
+                else:
+                    categories = util.getRequiredCategories(None)
+                self.parent.update_launcher_categories(categories, [])
+
+            if original is not None:
+                # Original found (this is a system-installed desktop file/
+                # directory rather than a user-made one), hide the desktop file
+                # and all associated instances in the menu
+                entry = MenulibreXdg.MenulibreDesktopEntry(original)
+                name = entry['Name']
+                comment = entry['Comment']
+                categories = entry['Categories']
+                executable = entry['Exec']
+                icon_name = entry['Icon']
+                hidden = entry['Hidden'] or entry['NoDisplay']
+                self.update_selected(name, comment, executable, categories,
+                                     item_type, icon_name, original, not hidden)
+                model, row_data = self.get_selected_row_data()
+                self.update_launcher_instances(filename, row_data)
+                treeiter = None
 
         if treeiter is not None:
             path = model.get_path(treeiter)
             if model is not None and treeiter is not None:
-                model.remove(treeiter)
+                if not isinstance(model, Gtk.TreeModelFilter):
+                    model.remove(treeiter)
             if path:
                 self._treeview.set_cursor(path)
 
         self.update_menus()
 
+    def remove_iter(self, model, treeiter):
+        """Remove launcher pointed to by iter from the model only - use this
+        when you need to remove a launcher from a non-selected directory, e.g.
+        when the associated category has been removed"""
+
+        # This feels a bit redundant for a function, but it keeps the
+        # functionality close to remove_selected
+        model.remove(treeiter)
+
 # Get
     def get_parent(self, model=None, treeiter=None):
         """Get the parent iterator for the current treeiter"""
@@ -212,16 +274,48 @@ class Treeview(GObject.GObject):
                 if path.get_depth() > 0:
                     try:
                         parent = model.get_iter(path)
-                    except:
+                    except:  # noqa
                         parent = None
         return model, parent
 
+    def is_first(self, model=None, treeiter=None):
+        if model is None:
+            model, treeiter = self._get_selected_iter()
+        if treeiter:
+            path = model.get_path(treeiter)
+            if path.prev():
+                return False
+        return True
+
+    def _next(self, model, treeiter, path):
+        # path.next() seems to be broken, so we will do it ourselves.
+        try:
+            string = path.to_string()
+            parts = string.split(":")
+            parts[-1] = str(int(parts[-1])+1)
+            string = ":".join(parts)
+            path = Gtk.TreePath.new_from_string(string)
+            model.get_iter(path)
+        except (TypeError, ValueError):
+            return None
+        return path
+
+    def is_last(self, model=None, treeiter=None):
+        if model is None:
+            model, treeiter = self._get_selected_iter()
+        if treeiter:
+            path = model.get_path(treeiter)
+
+            if self._next(model, treeiter, path) is not None:
+                return False
+        return True
+
     def get_parent_filename(self):
         """Get the filename of the parent iter."""
         model, parent = self.get_parent()
         if parent is None:
             return None
-        return model[parent][5]
+        return model[parent][MenuEditor.COL_FILENAME]
 
     def get_parent_row_data(self):
         """Get the row data of the parent iter."""
@@ -234,7 +328,7 @@ class Treeview(GObject.GObject):
         """Return the filename of the current selected treeiter."""
         model, row_data = self.get_selected_row_data()
         if row_data is not None:
-            return row_data[5]
+            return row_data[MenuEditor.COL_FILENAME]
         return None
 
     def get_selected_row_data(self):
@@ -255,28 +349,29 @@ class Treeview(GObject.GObject):
     def update_launcher_instances(self, filename, row_data):
         """Update all same launchers with the new information."""
         model, treeiter = self._get_selected_iter()
-        for instance in self._get_launcher_instances(model=model,
-                                                    filename=filename):
+        for instance in self._get_launcher_instances(filename, model):
             for i in range(len(row_data)):
                 model[instance][i] = row_data[i]
 
-    def update_selected(self, name, comment, item_type, icon_name, filename):
+    def update_selected(self, name, comment, executable, categories, item_type,
+                        icon_name, filename, show=True):
         """Update the application treeview selected row data."""
         model, treeiter = self._get_selected_iter()
-        model[treeiter][0] = name
-        model[treeiter][1] = comment
-        model[treeiter][2] = item_type
-
+        model[treeiter][MenuEditor.COL_NAME] = name
+        model[treeiter][MenuEditor.COL_COMMENT] = escapeText(comment)
+        model[treeiter][MenuEditor.COL_EXEC] = executable
+        model[treeiter][MenuEditor.COL_CATEGORIES] = categories
+        model[treeiter][MenuEditor.COL_TYPE] = item_type
         if os.path.isfile(icon_name):
             gfile = Gio.File.parse_name(icon_name)
             icon = Gio.FileIcon.new(gfile)
         else:
             icon = Gio.ThemedIcon.new(icon_name)
-        model[treeiter][3] = icon
+        model[treeiter][MenuEditor.COL_G_ICON] = icon
+        model[treeiter][MenuEditor.COL_ICON_NAME] = icon_name
+        model[treeiter][MenuEditor.COL_FILENAME] = filename
+        model[treeiter][MenuEditor.COL_SHOW] = show
 
-        model[treeiter][4] = icon_name
-        model[treeiter][5] = filename
-        
         # Refresh the displayed launcher
         self._last_selected_path = -1
         self._on_treeview_cursor_changed(self._treeview, None, None)
@@ -325,10 +420,10 @@ class Treeview(GObject.GObject):
         if self._toolbar.get_sensitive():
             model = treeview.get_model()
             row = model[treeiter]
-            row[6] = expanded
+            row[MenuEditor.COL_EXPAND] = expanded
 
     def _on_treeview_selection(self, sel, store, path, is_selected,
-                              can_select_func):
+                               can_select_func):
         """Save changes on cursor change."""
         if is_selected:
             return can_select_func()
@@ -350,23 +445,35 @@ class Treeview(GObject.GObject):
         expanded = self._get_treeview_selected_expanded(treeview)
         self._set_treeview_selected_expanded(treeview, not expanded)
 
+    def _text_display_func(self, col, renderer, treestore, treeiter,
+                           user_data):
+        """CellRenderer function to set the gicon for each row."""
+        show = treestore[treeiter][MenuEditor.COL_SHOW]
+        if show:
+            renderer.set_property("style", Pango.Style.NORMAL)
+        else:
+            renderer.set_property("style", Pango.Style.ITALIC)
+        renderer.set_property("style-set", True)
+
     def _icon_name_func(self, col, renderer, treestore, treeiter, user_data):
         """CellRenderer function to set the gicon for each row."""
-        renderer.set_property("gicon", treestore[treeiter][3])
+        renderer.set_property("gicon", treestore[treeiter][MenuEditor.COL_G_ICON])
 
     def _get_selected_iter(self):
         """Return the current treeview model and selected iter."""
         model, treeiter = self._treeview.get_selection().get_selected()
         return model, treeiter
 
-    def _populate_and_select_iter(self, model, treeiter, row_data):
-        """Fill the specified treeiter with data and select it."""
+    def _populate_and_select_iter(self, model, treeiter, row_data,
+                                  do_select=True):
+        """Fill the specified treeiter with data and optionally select it."""
         for i in range(len(row_data)):
             model[treeiter][i] = row_data[i]
 
-        # Select the new iter.
-        path = model.get_path(treeiter)
-        self._treeview.set_cursor(path)
+        # Select the new iter if requested
+        if do_select:
+            path = model.get_path(treeiter)
+            self._treeview.set_cursor(path)
 
     def _get_deletable_launcher(self, filename):
         """Return True if the launcher is available for deletion."""
@@ -374,18 +481,18 @@ class Treeview(GObject.GObject):
             return False
         return True
 
-    def _get_delete_filenames(self, model, treeiter):
+    def _get_delete_filenames(self, model, treeiter):  # noqa
         """Return a list of files to be deleted after uninstall."""
         directories = []
         applications = []
 
-        filename = model[treeiter][5]
+        filename = model[treeiter][MenuEditor.COL_FILENAME]
         block_run = False
 
         if filename is not None:
-            basename = getBasename(filename)
+            basename = getRelativeName(filename)
             original = util.getSystemLauncherPath(basename)
-            item_type = model[treeiter][2]
+            item_type = model[treeiter][MenuEditor.COL_TYPE]
             if original is None and item_type == MenuItemTypes.DIRECTORY:
                 pass
             else:
@@ -394,7 +501,7 @@ class Treeview(GObject.GObject):
         if model.iter_has_child(treeiter) and not block_run:
             for i in range(model.iter_n_children(treeiter)):
                 child_iter = model.iter_nth_child(treeiter, i)
-                filename = model[child_iter][5]
+                filename = model[child_iter][MenuEditor.COL_FILENAME]
                 if filename is not None:
                     if filename.endswith('.directory'):
                         d, a = self._get_delete_filenames(model, child_iter)
@@ -405,7 +512,7 @@ class Treeview(GObject.GObject):
                         if self._get_deletable_launcher(filename):
                             applications.append(filename)
 
-        filename = model[treeiter][5]
+        filename = model[treeiter][MenuEditor.COL_FILENAME]
         if filename is not None:
             if filename.endswith('.directory'):
                 directories.append(filename)
@@ -421,23 +528,23 @@ class Treeview(GObject.GObject):
         row = model[treeiter]
         return treeview.row_expanded(row.path)
 
-    def _get_launcher_instances(self, model=None, parent=None, filename=None):
+    def _get_launcher_instances(self, filename, model=None, parent=None):
         """Return a list of all treeiters referencing this filename."""
         if model is None:
             model, treeiter = self._get_selected_iter()
         treeiters = []
         for n_child in range(model.iter_n_children(parent)):
             treeiter = model.iter_nth_child(parent, n_child)
-            iter_filename = model[treeiter][5]
+            iter_filename = model[treeiter][MenuEditor.COL_FILENAME]
             if iter_filename == filename:
                 treeiters.append(treeiter)
             if model.iter_has_child(treeiter):
-                treeiters += self._get_launcher_instances(model, treeiter,
-                                                         filename)
+                treeiters += self._get_launcher_instances(filename, model,
+                                                          treeiter)
         return treeiters
 
     def _get_n_launcher_instances(self, filename):
-        return len(self._get_launcher_instances(filename=filename))
+        return len(self._get_launcher_instances(filename))
 
     def _is_menu_locked(self):
         """Return True if menu editing is currently locked."""
@@ -455,7 +562,7 @@ class Treeview(GObject.GObject):
         treestore, treeiter = self._get_selected_iter()
         model, parent_iter = self.get_parent()
         while parent_iter is not None:
-            filename = treestore[parent_iter][5]
+            filename = treestore[parent_iter][MenuEditor.COL_FILENAME]
             if getBasename(filename).startswith(prefix):
                 add_enabled = False
             model, parent_iter = self.get_parent(treestore, parent_iter)
@@ -507,14 +614,14 @@ class Treeview(GObject.GObject):
                 for n_child in range(model.iter_n_children(None)):
                     treeiter = model.iter_nth_child(None, n_child)
                     row = model[treeiter]
-                    if row[6]:
+                    if row[MenuEditor.COL_FILENAME]:
                         self._treeview.expand_row(row.path, False)
 
                 # Try to get the row that was selected previously.
                 if (f_model is not None) and (f_iter is not None):
                     row_data = f_model[f_iter][:]
                     selected_iter = self._get_iter_by_data(row_data, model,
-                                                          parent=None)
+                                                           parent=None)
 
                 # If that fails, just select the first iter.
                 else:
@@ -526,8 +633,8 @@ class Treeview(GObject.GObject):
 
     def _treeview_match(self, model, treeiter, query):
         """Match subfunction for filtering search results."""
-        name, comment, item_type, icon, pixbuf, desktop, expanded = \
-                model[treeiter][:]
+        name, comment, executable, categories, item_type, icon, pixbuf, desktop, \
+            expanded, show = model[treeiter][:]
 
         # Hide separators in the search results.
         if item_type == MenuItemTypes.SEPARATOR:
@@ -538,6 +645,8 @@ class Treeview(GObject.GObject):
             name = ""
         if not comment:
             comment = ""
+        if not executable:
+            executable = ""
 
         # Expand all the rows.
         self._treeview.expand_all()
@@ -550,6 +659,15 @@ class Treeview(GObject.GObject):
         if query in comment.lower():
             return True
 
+        if query in executable.lower():
+                return True
+
+        # Match against the desktop file.
+        desktop = desktop.replace("menulibre-", "")
+        desktop = desktop.replace("alacarte-", "")
+        if query in desktop.lower():
+            return True
+
         # Show the directory if any child items match.
         if item_type == MenuItemTypes.DIRECTORY:
             return self._treeview_match_directory(query, model, treeiter)
@@ -582,20 +700,20 @@ class Treeview(GObject.GObject):
             return
         if filename.endswith('.desktop'):
             menu_install = True
-            menu_prefix = util.getDefaultMenuPrefix()
             parents = []
             if parent is None:
                 parent = model.iter_parent(treeiter)
             while parent is not None:
-                parent_filename = model[parent][5]
+                parent_filename = model[parent][MenuEditor.COL_FILENAME]
+
                 # Do not do this method if this is a known system directory.
-                if getBasename(parent_filename).startswith(menu_prefix):
-                    menu_install = False
                 parents.append(parent_filename)
                 parent = model.iter_parent(parent)
             parents.reverse()
             if menu_install:
-                MenulibreXdg.desktop_menu_install(parents, [filename])
+                installed = MenulibreXdg.desktop_menu_install(parents, [filename])
+                if not installed:
+                    self.emit("requires-menu-reload", True)
 
     def xdg_menu_uninstall(self, model, treeiter, filename):
         """Uninstall the specified filename from the menu structure."""
@@ -607,7 +725,7 @@ class Treeview(GObject.GObject):
             parents = []
             parent = model.iter_parent(treeiter)
             while parent is not None:
-                parent_filename = model[parent][5]
+                parent_filename = model[parent][MenuEditor.COL_FILENAME]
                 # Do not do this method if this is a known system directory.
                 if getBasename(parent_filename).startswith(menu_prefix):
                     menu_install = False
@@ -623,20 +741,37 @@ class Treeview(GObject.GObject):
 
     def update_menus(self):
         """Update the menu files."""
+        if self.menu_timeout_id > 0:
+            GLib.source_remove(self.menu_timeout_id)
+        self.menu_timeout_id = GLib.timeout_add_seconds(1,
+            self.update_menu_timeout)
+
+    def update_menu_timeout(self):
         # Do not save menu layout if in search mode (lp #1306999)
         if not self._is_menu_locked():
             XmlMenuElementTree.treeview_to_xml(self._treeview)
+            self.update_menus_kde()
+        self.menu_timeout_id = 0
+        return False
+
+    def update_menus_kde(self):
+        try:
+            subprocess.Popen(["kbuildsycoca5"])
+        except FileNotFoundError:
+            pass
 
     def _cleanup_applications_merged(self):
         """Cleanup items from ~/.config/menus/applications-merged"""
         # xdg-desktop-menu installs menu files in
-        # ~/.config/menus/applications-merged, but does remove them correctly.
+        # ~/.config/menus/applications-merged, but does not remove them
+        # correctly.
         merged_dir = os.path.join(GLib.get_user_config_dir(),
-                                    "menus", "applications-merged")
+                                  "menus", "applications-merged")
 
         # Get the list of installed user directories to compare with.
         directories_dir = os.path.join(GLib.get_home_dir(),
-            ".local", "share", "desktop-directories")
+                                       ".local", "share",
+                                       "desktop-directories")
         if os.path.isdir(directories_dir):
             directories = os.listdir(directories_dir)
         else:
@@ -651,7 +786,7 @@ class Treeview(GObject.GObject):
                 # Only interested in .menu files
                 if os.path.isfile(menufile) and menufile.endswith('.menu'):
                     logger.debug("Checking if %s is still valid..." %
-                                menufile)
+                                 menufile)
 
                     # Read the menufile to see if it has a valid directory.
                     with open(menufile) as menufile_tmp:
@@ -661,7 +796,7 @@ class Treeview(GObject.GObject):
                                 menuname = menuname.split('</Directory>')[0]
                                 menuname = menuname.strip()
 
-                                # If a listed directory is not installed, remove
+                                # Remove if a listed directory is not installed
                                 if menuname not in directories:
                                     remove_file = True
                     if remove_file:
@@ -669,7 +804,7 @@ class Treeview(GObject.GObject):
                         os.remove(menufile)
 
 # TreeView iter tricks
-    def _move_iter(self, widget, user_data):
+    def _move_iter(self, widget, user_data):  # noqa
         """Move the currently selected row up or down. If the neighboring row
         is expanded, make the selected row a child of the neighbor row.
 
@@ -687,12 +822,12 @@ class Treeview(GObject.GObject):
         sel = treeview.get_selection().get_selected()
         if sel:
             model, selected_iter = sel
-            selected_type = model[selected_iter][2]
+            selected_type = model[selected_iter][MenuEditor.COL_TYPE]
 
             # Get current required categories
             model, parent = self.get_parent(model, selected_iter)
             if parent:
-                categories = util.getRequiredCategories(model[parent][5])
+                categories = util.getRequiredCategories(model[parent][MenuEditor.COL_FILENAME])
             else:
                 categories = util.getRequiredCategories(None)
 
@@ -709,7 +844,7 @@ class Treeview(GObject.GObject):
                 move_down = False
 
                 # What is the neighboring item?
-                sibling_type = model[sibling_iter][2]
+                sibling_type = model[sibling_iter][MenuEditor.COL_TYPE]
 
                 # Sibling Directory
                 if sibling_type == MenuItemTypes.DIRECTORY:
@@ -727,9 +862,9 @@ class Treeview(GObject.GObject):
 
                 # Insert the selected item into the directory.
                 if move_down:
-                    selected_iter = self._move_iter_down_level(treeview,
-                                        selected_iter,
-                                        sibling_iter, relative_position)
+                    selected_iter = self._move_iter_down_level(
+                        treeview, selected_iter,
+                        sibling_iter, relative_position)
 
                 # Move the selected item before or after the sibling item.
                 else:
@@ -740,14 +875,13 @@ class Treeview(GObject.GObject):
 
             # If there is no neighboring row, move up a level.
             else:
-                selected_iter = self._move_iter_up_level(treeview,
-                                        selected_iter,
-                                        relative_position)
+                selected_iter = self._move_iter_up_level(
+                    treeview, selected_iter, relative_position)
 
             # Get new required categories
             model, parent = self.get_parent(model, selected_iter)
             if parent:
-                new_categories = util.getRequiredCategories(model[parent][5])
+                new_categories = util.getRequiredCategories(model[parent][MenuEditor.COL_FILENAME])
             else:
                 new_categories = util.getRequiredCategories(None)
 
@@ -764,10 +898,15 @@ class Treeview(GObject.GObject):
                 split_categories.sort()
                 editor_categories = ';'.join(split_categories)
                 self.parent.set_editor_categories(editor_categories)
-                self.parent.update_launcher_categories(categories, new_categories)
+                self.parent.update_launcher_categories(categories,
+                                                       new_categories)
 
         self.update_menus()
 
+        self.scroll_to_selection()
+
+        self.emit("cursor-changed", True)
+
     def _get_iter_by_data(self, row_data, model, parent=None):
         """Search the TreeModel for a row matching row_data.
 
@@ -799,7 +938,7 @@ class Treeview(GObject.GObject):
                                               row_data)
 
             # Install/Uninstall items from directories.
-            filename = row_data[5]
+            filename = row_data[MenuEditor.COL_FILENAME]
             self.xdg_menu_install(filename)
             self.xdg_menu_uninstall(model, treeiter, filename)
 
@@ -809,7 +948,7 @@ class Treeview(GObject.GObject):
             return new_iter
 
     def _move_iter_down_level(self, treeview, treeiter, parent_iter,
-                             relative_position):
+                              relative_position):
         """Move the specified iter down one level."""
         model = treeview.get_model()
         row_data = model[treeiter][:]
@@ -825,7 +964,7 @@ class Treeview(GObject.GObject):
             new_iter = model.insert(parent_iter, 0, row_data)
 
         # Install/Uninstall items from directories.
-        filename = row_data[5]
+        filename = row_data[MenuEditor.COL_FILENAME]
         self.xdg_menu_install(filename, parent_iter)
         self.xdg_menu_uninstall(model, treeiter, filename)
 
@@ -834,3 +973,50 @@ class Treeview(GObject.GObject):
         path = model.get_path(new_iter)
         treeview.set_cursor(path)
         return new_iter
+
+    def _sort_iter(self, widget):
+        """Alphabetical sort of items in the current directory."""
+
+        # Get the current selected row
+        model, sel_iter = self._get_selected_iter()
+        if sel_iter:
+
+            # Move to the parent iter - if there is no parent, it must be the
+            # top level, which is ignored
+            item_names = []
+            _, parent_iter = self.get_parent(model, sel_iter)
+            if parent_iter:
+
+                # Deteriming list of item names
+                for i in range(model.iter_n_children(parent_iter)):
+                    child_iter = model.iter_nth_child(parent_iter, i)
+                    item_names.append(model[child_iter][MenuEditor.COL_NAME])
+
+                # Applying unstable (?) case-insensitive alphabetical sort
+                item_names = sorted(item_names, key=str.lower)
+
+                for i in range(len(item_names)):
+                    child_iter = model.iter_nth_child(parent_iter, i)
+
+                    # Ignore if item is already sorted or at least has an
+                    # identical title to that expected
+                    if item_names[i] != model[child_iter][MenuEditor.COL_NAME]:
+
+                        # Locating desired item in the remaining unsorted items
+                        for r in range(i, len(item_names)):
+                            search_iter = model.iter_nth_child(parent_iter, r)
+                            if item_names[i] == model[search_iter][MenuEditor.COL_NAME]:
+                                break
+
+                        # Moving the found item into place
+                        model.move_before(search_iter, child_iter)
+
+                # Committing changes
+                self.update_menus()
+
+            self.scroll_to_selection()
+
+    def scroll_to_selection(self):
+        model, sel_iter = self._get_selected_iter()
+        self._treeview.scroll_to_cell(model.get_path(sel_iter), None,
+                                False, 0.0, 0.0)
\ No newline at end of file
